#!/usr/bin/python3
""" Takes web log file and replaces ips with a number to allows for anonymized analysis.
    Optionally some ips and usernames can be stripped also.
"""
import argparse
import netaddr
import re
import sys


def anonymize(new_file, log_file, skip_ips=None, skip_private=False, strip_usernames=False, username_field=2):
    """
    :param new_file: A file object that will be written out anonymized
    :param log_file: A file object to read from
    :param skip_ips: A netaddr.IPSet of ips to skip
    :param skip_private: A boolean, true to skip private ips
    :param strip_usernames: A boolean, true to strip usernames
    :param username_field: An integer specifying the field usernames are indexed by
    :return: (total lines processed, lines skipped)
    """
    ips = {}
    total = 0
    skip = 0
    for line in log_file:
        total += 1
        splits = line.split()
        ip_raw = splits[0].decode('utf-8')
        # Work around nginx general access/error logs that emit the date instead of IP as the first column.
        if re.match('^\d+/\d+/\d+', ip_raw):
            skip += 1
            continue
        try:
            ip = netaddr.IPAddress(ip_raw)
        except netaddr.AddrFormatError:
            skip += 1
            continue

        if (skip_private and ip.is_private()) or (skip_ips is not None and ip in skip_ips):
            skip += 1
            continue

        if not str(ip) in ips:
            ips[str(ip)] = str(len(ips))

        if strip_usernames:
            del splits[username_field]

        anonymized_line = "{} {}\n".format(ips[str(ip)], b' '.join(splits[1:]).decode('utf-8'))
        new_file.write(anonymized_line.encode('utf-8'))

    return total, skip


def parse_args(argv):
    # Parse arguments, build skip_set and paths as needed
    parser = argparse.ArgumentParser()
    parser.add_argument('log_path', help='The path to the log to be anonymized')
    parser.add_argument('--output', '-o', help='Output file, defaults to input file'
                                               'with .anonymized extension')
    parser.add_argument('--skip', '-s', action='append', help='A cidr notation network for which any'
                                                              'for which any source traffic will be removed'
                                                              'from the final output. Can be specified multiple'
                                                              'times')
    parser.add_argument('--skip_private', action='store_true', default=False,
                        help='If specified remove any private source ips from the final file')

    parser.add_argument('--strip_usernames', action='store_true', default=False,
                        help='Strip usernames from log lines.')
    parser.add_argument('--username_field', type=int, default=2,
                        help='The username index from the space separated list of words on each log line.')

    parser.add_argument('--verbose', '-v', action='store_true', default=False,
                        help='Verbose: output lines processed/skipped')
    args = parser.parse_args(argv)

    if args.skip is not None:
        args.skip_ips = netaddr.IPSet()
        for cidr in args.skip:
            args.skip_ips.add(cidr)
    else:
        args.skip_ips = None

    return args


def main():
    args = parse_args(sys.argv[1:])
    if args.output is None:
        anonymized_path = args.log_path + '.anonymized'
    else:
        anonymized_path = args.output

    # Process the lines of the file
    with open(anonymized_path, 'w') as new_file:
        with open(args.log_path, 'r') as log_file:
            anonymize_args = (
                new_file,
                log_file,
                args.skip_ips,
                args.skip_private,
                args.strip_usernames,
                args.username_field
            )
            total, skip = anonymize(*anonymize_args)

    if args.verbose:
        print(('Total lines processed {}, skipped {}'.format(total, skip)))


if __name__ == "__main__":
    sys.exit(main())
