#!/usr/bin/python3
#
# -------------------------------------------------------#
#  This file is Juju managed - do not make local changes #
# -------------------------------------------------------#
#
# Author: Chris Stratford <chris.stratford@canonical.com>
# Copyright: Canonical Ltd. 2016
#
# Archive log files to the Azure blobstore
#
# Should be quiet on success, otherwise reports any errors to stdout
# Useful info logged to a status file (file overwritten each run)
#
# The standard invocation setup by the charm is
# 'archive-to-azure.py -s -f [config_file] -l [logdir1] -l [logdir2] -c [container_name] -a [archive_after] -d [delete_after] -p [retention_days]'

import os
import logging
import sys
import datetime

import httplib2
import json
import hmac
import hashlib
import base64
import xml.etree.ElementTree as ET
from archive import CloudStorage, main, Status

log = logging.getLogger()


class AzureCloudStorage(CloudStorage):
    """ A class wrapping the interface to the Azure blob store.
        The intent is to hide the complexity of login and other common operations.
        bucket = container
        project = storage account
        API details at - https://msdn.microsoft.com/en-GB/library/azure/dd135733.aspx
    """
    def __init__(self, json_file, bucket, project=None):
        creds_file = json.loads(open(json_file, 'r').read())
        self.primary_key = creds_file["private_key"]
        self.uri = "https://{}.blob.core.windows.net/".format(project)
        self.json_file = json_file
        self.bucket, self.bucket_path = bucket.split('/', 1)
        self.maxfile = 64 * 1024 * 1024
        # Create the bucket if it doesn't exist and a project is defined
        self.project = project
        if self.project is not None:
            self._create_bucket()
        else:  # If there is no possibility of creating the bucket error early if it doesn't exist
            self.info()

    def _calculate_str_md5(self, s, style="base64"):
        """ Calculate and return an md5 for a string.
        :param s: The string to calculate an md5 for
        :param style: The output format, one of "base64" (default) or "hexdigest"
        :return: The md5 hash
        """
        return super(AzureCloudStorage, self)._calculate_str_md5(s, style=style)

    def standard_timestamp(self, timestamp):
        """ Convert a GCE timestamp to something standardised
        Azure returns a timestamp like "Thu, 25 Feb 2016 17:36:11 GMT"
        :param timestamp: A file timestamp in GCE format
        :return: timestamp in standardised format
        """
        return datetime.datetime.strptime(timestamp, '%a, %d %b %Y %H:%M:%S %Z')

    def _dict_to_str(self, d, kv_sep=":", entry_sep="\n", strip_final=True, startswith=""):
        s = ""
        for key in sorted(d):
            if key.startswith(startswith):
                s += key + kv_sep + d[key] + entry_sep
        if strip_final:
            s = s.rstrip(entry_sep)
        return s

    def _get_ms_date(self, d=None):
        if not d:
            d = datetime.datetime.now()
        ms_date = d.strftime('%a, %d %b %Y %H:%M:%S %Z')
        # Force GMT if no TZ provided
        if not d.strftime('%Z'):
            ms_date += "GMT"
        return ms_date

    def _create_azure_auth_signature(self, container, headers, query, method):
        # Note: auth_headers are in a specific order
        # see https://msdn.microsoft.com/en-gb/library/azure/dd179428.aspx
        auth_headers = ("Content-Encoding", "Content-Language", "Content-Length", "Content-MD5",
                        "Content-Type", "Date", "If-Modified-Since", "If-Match", "If-None-Match",
                        "If-Unmodified-Since", "Range")
        canonicalized_headers = self._dict_to_str(headers, strip_final=False, startswith="x-ms-")
        query_string = self._dict_to_str(query)

        canonicalized_resource = "/{}/{}".format(self.project, container)
        if query_string:
            canonicalized_resource += "\n" + query_string

        # Create the auth signature string
        signature = method + "\n"
        for header in auth_headers:
            signature += headers.get(header, "") + "\n"
        signature += canonicalized_headers
        signature += canonicalized_resource

        encoded_sig = base64.b64encode(hmac.new(base64.b64decode(self.primary_key),
                                                msg=str(signature, "utf-8"),
                                                digestmod=hashlib.sha256).digest())
        return encoded_sig

    # Requests documented here: https://msdn.microsoft.com/en-GB/library/azure/dd135733.aspx
    def _send_azure_request(self, container, query={}, headers={}, body=None, method="GET"):
        httplib2.debuglevel = 0
        h = httplib2.Http()

        if method == "PUT" and "Content-Length" not in headers:
            headers["Content-Length"] = "0"
        headers["x-ms-date"] = self._get_ms_date()
        headers["x-ms-version"] = "2009-09-19"
        # headers["x-ms-client-request-id"] = "chriss-test-1"

        auth = self._create_azure_auth_signature(container, headers, query, method)
        headers["Authorization"] = "SharedKey {}:{}".format(self.project, auth)

        query_string = self._dict_to_str(query, kv_sep="=", entry_sep="&")
        response, content = h.request(uri=self.uri + container + "?" + query_string,
                                      method=method,
                                      body=body,
                                      headers=headers)
        return response, content

    def _create_bucket(self):
        """ Creates the bucket if it doesn't yet exist. A project must be specified for this to work.
        """
        log.info('Creating bucket {}'.format(self.bucket))
        response, content = self._send_azure_request(self.bucket,
                                                     query={"restype": "container"},
                                                     headers={},
                                                     method="PUT")
        if response.get("status") == "201":
            log.info('Bucket {} created'.format(self.bucket))
            return Status.SUCCESS
        elif response.get("status") == "409":
            log.info("Bucket {} already exists".format(self.bucket))
            return Status.SUCCESS
        else:
            log.info("Bucket creation failed: {} {}".format(response.get("status"), response.get("reason")))
            return Status.FAILURE

    def upload_large_file(self, path):
        """ A direct push of the file (>64MB) at path to the cloud storage, no md5 compare just overwrite
        :param path: The local path of the file to upload
        """
        log.info('Uploading {} to Azure storage (large file)'.format(path))
        max_block = 4194304     # 4MB
        block_list = []
        block_number = 0
        file_name = os.path.basename(path)

        try:
            with open(path, "rb") as f:
                block_content = f.read(max_block)
                while block_content:
                    # Block_id must be the same size for all blocks
                    block_id = base64.b64encode("{:06}".format(block_number))
                    headers = {}
                    headers["Content-Length"] = str(len(block_content))
                    headers["Content-MD5"] = self._calculate_str_md5(block_content)
                    log.debug("Uploading block {}  size: {} id {}".format(block_number, len(block_content), block_id))
                    response, content = self._send_azure_request(self.bucket + "/" + file_name,
                                                                 query={"comp": "block", "blockid": block_id},
                                                                 headers=headers,
                                                                 body=block_content,
                                                                 method="PUT")
                    if not response.get("status") == "201":
                        log.info("Upload failed: {}".format(response.get("reason")))
                        return Status.FAILURE

                    block_list.append(block_id)
                    block_content = f.read(max_block)
                    block_number += 1
        except IOError:
            log.info("Unable to read {}".format(path))
            return Status.FAILURE

        # Now update the blocklist
        log.debug("Updating the blocklist")
        xml = '<?xml version="1.0" encoding="utf-8"?>' + "\n"
        xml += "<BlockList>\n"
        for block_id in block_list:
            xml += "  <Uncommitted>" + block_id + "</Uncommitted>\n"
        xml += "</BlockList>\n"
        headers = {}
        headers["Content-Length"] = str(len(xml))

        response, content = self._send_azure_request(self.bucket + "/" + self.expand_path(file_name),
                                                     query={"comp": "blocklist"},
                                                     headers=headers,
                                                     body=xml,
                                                     method="PUT")
        if response.get("status") == "201":
            log.debug("Blob {} uploaded".format(file_name))
            return Status.SUCCESS
        else:
            log.info("Blob upload failed: {}".format(response.get("reason")))
            return Status.FAILURE

    def upload_file(self, path):
        """ A direct push of the file at path to the cloud storage, no md5 compare just overwrite
        :param path: The local path of the file to upload
        """
        file_name = os.path.basename(path)
        file_size = os.stat(path).st_size
        if file_size > self.maxfile:
            return self.upload_large_file(path)
        log.info('Uploading {} to Azure storage'.format(path))

        content = ""
        try:
            with open(path, "rb") as f:
                content = f.read()
        except IOError:
            log.info("Unable to read {}".format(path))
            return Status.FAILURE

        headers = {}
        headers["Content-Length"] = str(len(content))
        # headers["Content-Type"] = "text/plain"
        headers["x-ms-blob-type"] = "BlockBlob"

        headers["Content-MD5"] = self._calculate_str_md5(content)
        response, content = self._send_azure_request(self.bucket + "/" + self.expand_path(file_name),
                                                     headers=headers,
                                                     body=content,
                                                     method="PUT")
        if response.get("status") == "201":
            log.debug("Blob {} uploaded".format(file_name))
            return Status.SUCCESS
        else:
            log.info("status: {} reason: {} content: {}".format(response.get("status"), response.get("reason"), content))
            return Status.FAILURE

    def delete(self, name):
        """ Delete an object from cloud storage
        :param name: The name of the object to delete
        :return:
        """
        log.info('Deleting object {} from bucket {}'.format(name, self.bucket))
        response, content = self._send_azure_request(self.bucket + "/" + self.expand_path(name),
                                                     method="DELETE")
        if response.get("status") == "202":
            log.info('Object {} deleted'.format(name))
            return Status.SUCCESS
        log.info("Unable to delete object: {} {}".format(response.get("status"), response.get("reason")))
        return Status.FAILURE

    def info(self, name=None):
        """ Get information on an object
        Returned object information must contain the following fields:
        name, size, contentType, updated, md5Hash
        :param name: If name is None, provide bucket information otherwise object information
        :return: A dictionary representation of the bucket or object information.
        """
        if not name:
            # Bucket information (TODO)
            return {}

        # Object information
        for obj in self.list_objects(self.expand_path(name)):
            if obj['name'] == self.expand_path(name):
                return obj

    def list_objects(self, name=None):
        """ List objects in self.bucket, this lists all objects even if multiple requests to the
            api are required to retreive them all.
            Azure doesn't have timeCreated, so we'll assume same as Last-Modified
        :param name: If name is none list all objects in the container
        :return: A list of dictionaries with these fields name, size, contentType, timeCreated, updated, md5Hash.
                  the list is empty is nothing is found.
        """
        obj_list = []
        query = {"restype": "container", "comp": "list", "maxresults": "5000"}
        query['prefix'] = self.expand_path(name)
        nextmarker = 999
        while nextmarker:
            response, content = self._send_azure_request(self.bucket, query=query)
            if response.get("status") == "200":
                root = ET.fromstring(content)
                blobs = root.find("Blobs")
                for blob in blobs.findall("Blob"):
                    blob_name = blob.find("Name").text
                    if not name or name == blob_name:
                        obj = {}
                        properties = blob.find("Properties")
                        obj["name"] = blob_name
                        obj["size"] = int(properties.find("Content-Length").text)
                        obj["contentType"] = properties.find("Content-Type").text
                        obj["timeCreated"] = properties.find("Last-Modified").text
                        obj["updated"] = properties.find("Last-Modified").text
                        obj["md5Hash"] = properties.find("Content-MD5").text
                        obj_list.append(obj)
                nextmarker = root.find("NextMarker").text
                if nextmarker:
                    query["marker"] = nextmarker

        return obj_list


if __name__ == "__main__":
    sys.exit(main(AzureCloudStorage))
