#!/usr/bin/python3
#
# -------------------------------------------------------#
#  This file is Juju managed - do not make local changes #
# -------------------------------------------------------#
#
# Author: Tim Kuhlman <timothy.kuhlman@canonical.com>
# Copyright: Canonical Ltd. 2015
#
# Archive log files to Google Compute Engine Storage
#
# Should be quiet on success, otherwise reports any errors to stdout
# Useful info logged to a status file (file overwritten each run)
#
# The standard invocation setup by the charm is
# 'archive-to-gce.py -s -f [config_file] -l [logdir1] -l [logdir2] -c [container_name] -a [archive_after] -d [delete_after] -p [retention_days]'

import os
import io
import json
import logging
import sys
import datetime

import httplib2
import apiclient
from apiclient import discovery
from apiclient.http import MediaIoBaseUpload
from oauth2client.service_account import ServiceAccountCredentials
from archive import CloudStorage, main, Status

log = logging.getLogger()


class GCECloudStorage(CloudStorage):
    """ A class wrapping the interface to Google Compute Engine Cloud storage.
        The intent is to hide the complexity of login and other common operations.
        API details at - https://cloud.google.com/storage/docs/json_api/v1/?hl=en_US
    """
    def __init__(self, json_file, bucket, project=None):
        self.json_file = json_file
        self.bucket, self.bucket_path = bucket.split('/', 1)
        self._gce_login()
        # Create the bucket if it doesn't exist and a project is defined
        self.project = project
        if self.project is not None:
            self._create_bucket()
        else:  # If there is no possibility of creating the bucket error early if it doesn't exist
            self.info()

    def standard_timestamp(self, timestamp):
        """ Convert a GCE timestamp to something standardised
        GCE returns a timestamp like "2015-10-15T20:45:10.583Z"
        :param timestamp: A file timestamp in GCE format
        :return: timestamp in standardised format
        """
        return datetime.datetime.strptime(timestamp, '%Y-%m-%dT%H:%M:%S.%fZ')

    def _create_bucket(self):
        """ Creates the bucket if it doesn't yet exist. A project must be specified for this to work.
            Other than the name all options are default, ie location is US, etc. To change parameters for a
            bucket it is best to create it before running this script.
        """
        # Create the bucket
        req = self.service.buckets().insert(project=self.project, body={'name': self.bucket})
        try:
            log.debug('Creating bucket {}'.format(self.bucket))
            resp = req.execute()
        except apiclient.errors.HttpError as e:
            # If the bucket already exists continue
            if e.resp.status == 409:
                log.debug('Bucket {} already exists'.format(self.bucket))
            else:
                raise
        else:
            log.debug(json.dumps(resp, indent=2))

    def _gce_login(self):
        # The ServiceAccountCredentials loads creds directly from the json file
        # and creats creds with a specific set of scopes as defined by the second
        # parameter.  
        creds = ServiceAccountCredentials.from_json_keyfile_name(self.json_file, [ 'https://www.googleapis.com/auth/devstorage.read_write' ])
        http = creds.authorize(httplib2.Http())
        self.service = discovery.build('storage', 'v1', http=http)

    def upload_file(self, path):
        """ A direct push of the file at path to the cloud storage, no md5 compare just overwrite
        :param path: The local path of the file to upload
        """
        file_name = os.path.basename(path)
        log.info('Uploading {} to GCE storage'.format(file_name))
        try:
            media = MediaIoBaseUpload(io.FileIO(path, 'r'), 'text/plain')
        except IOError:
            log.info("Unable to read {}".format(path))
            return Status.FAILURE
        req = self.service.objects().insert(bucket=self.bucket, name=self.expand_path(file_name), media_body=media)
        log.debug('Push output {}'.format(req.execute()))
        # FIXME - test status
        return Status.SUCCESS

    def delete(self, name):
        """ Delete an object from cloud storage
        :param name: The name of the object to delete
        :return:
        """
        log.info('Deleting object {} from bucket {}'.format(self.expand_path(name), self.bucket))
        return self.service.objects().delete(bucket=self.bucket, object=self.expand_path(name)).execute()

    def info(self, name=None):
        """ Get information about an object
        Returned object information must contain the following fields:
        name, size, contentType, updated, md5Hash
        :param name: If name is None, provide bucket information otherwise object information
        :return: A dictionary representation of the bucket or object information.
        """
        if not name:
            # Bucket information
            try:
                req = self.service.buckets().get(bucket=self.bucket)
                resp = req.execute()
            except apiclient.errors.HttpError as e:
                if e.resp.status == 404:
                    log.debug('Bucket {} does not exist'.format(self.bucket))
                raise
            else:
                return resp
        # Object information
        for obj in self.list_objects(name):
            if self.strip_path(obj['name']) == name:
                return obj

    def list_objects(self, name=None):
        """ List objects in self.bucket, this lists all objects even if multiple requests to the api are required to
            retreive them all.
        :param name: If name is none list all objects in the container
        :return: A list of dictionaries with these fields name, size, contentType, timeCreated, updated, md5Hash.
                  the list is empty is nothing is found.
        """
        # Create a request to objects.list to retrieve a list of objects.
        fields_to_return = 'nextPageToken,items(name,size,contentType,timeCreated,updated,md5Hash)'
        req = self.service.objects().list(bucket=self.bucket,
                                          fields=fields_to_return,
                                          prefix=self.expand_path(name))

        objects = []
        # If you have too many items to list in one request, list_next() will
        # automatically handle paging with the pageToken.
        while req:
            resp = req.execute()
            if len(resp) == 0:
                return objects
            for item in resp['items']:
                if 'size' in item:
                    item['size'] = int(item['size'])
                objects.append(item)
            req = self.service.objects().list_next(req, resp)

        return objects


if __name__ == "__main__":
    sys.exit(main(GCECloudStorage))
