#!/usr/bin/python3
#
# -------------------------------------------------------#
#  This file is Juju managed - do not make local changes #
# -------------------------------------------------------#
#
# Author: Chris Stratford <chris.stratford@canonical.com>
# Copyright: Canonical Ltd. 2014
#
# Archive log files to S3 storage
#
# Should be quiet on success, otherwise reports any errors to stdout
# Useful info logged to a status file (file overwritten each run)
#
# The standard invocation setup by the charm is
# 'archive-to-s3.py -s -f [config_file] -l [logdir1] -l [logdir2] -c [container_name] -a [archive_after] -d [delete_after] -p [retention_days]'

import os
import logging
import sys
import datetime

import json
import boto
from boto.s3.connection import OrdinaryCallingFormat
from archive import CloudStorage, main, Status

log = logging.getLogger()


class S3CloudStorage(CloudStorage):
    """ A class wrapping the interface to S3 storage
        The intent is to hide the complexity of login and other common operations.
        bucket = container
        project = storage account
    """
    def __init__(self, json_file, bucket, project=None):
        creds = json.loads(open(json_file, 'r').read())
        wanted = ["AWS_ACCESS_KEY_ID",
                  "AWS_SECRET_ACCESS_KEY",
                  "AWS_REGION"]
        config = {}
        for config_name in wanted:
            if config_name not in creds:
                # Missing creds
                log.info("Missing credentials ({}) in config file".format(config_name))
                raise KeyError
            config[config_name] = creds[config_name]

        self.connect(config)

        self.bucket, self.bucket_path = bucket.split('/', 1)
        self.project = project
        self.segment_size = 1073741824  # Must be >5MB and <5GB

        # Ensure the bucket exists before using it
        try:
            if config['AWS_REGION'] == 'us-east-1':
                # us-east-1 is a special case
                self.s3bucket = self.s3.create_bucket(self.bucket)
            else:
                self.s3bucket = self.s3.create_bucket(self.bucket, location=config['AWS_REGION'])
        except boto.exception.S3CreateError:
            self.s3bucket = self.s3.get_bucket(self.bucket)

        # Make sure we're in the same region as the bucket
        # and reconnect in the right one if not
        for bucket in self.s3.get_all_buckets():
            if bucket.name == self.bucket:
                websiteEndpoint = bucket.get_website_endpoint()
                break
        websiteEndpoint = websiteEndpoint[websiteEndpoint.find(".") + 1:]

        # Find the real region
        for region in boto.regioninfo.get_regions("s3"):
            if region.name in websiteEndpoint:
                newRegion = region.name
                if newRegion != config['AWS_REGION']:
                    log.info("Reconnecting to endpoint in {}".format(newRegion))
                    config["AWS_REGION"] = newRegion
                    self.connect(config)

    def connect(self, config):
        for region in boto.regioninfo.get_regions("s3"):
            if region.name == config['AWS_REGION']:
                endpoint = region.endpoint

        self.s3 = boto.s3.connection.S3Connection(config['AWS_ACCESS_KEY_ID'],
                                                  config['AWS_SECRET_ACCESS_KEY'],
                                                  is_secure=True,
                                                  host=endpoint,
                                                  calling_format=OrdinaryCallingFormat())

    def _calculate_md5(self, path, style="hexdigest"):
        """ Calculate and return an md5 for a local file.
        :param path: The local filesystem path to calculate an md5 for
        :param style: The output format, one of "base64" or "hexdigest" (default)
        :return: The md5 hash
        """
        return super(S3CloudStorage, self)._calculate_md5(path, style=style)

    def standard_timestamp(self, timestamp):
        """ Convert a GCE timestamp to something standardised
        S3 returns a timestamp like "2016-03-11T15:26:05.000Z" when listing objects
        and "Fri, 11 Mar 2016 15:26:05 GMT" when stat-ing them, so we handle both
        :param timestamp: A file timestamp in one of S3's formats
        :return: timestamp in standardised format
        """
        try:
            d = datetime.datetime.strptime(timestamp, '%Y-%m-%dT%H:%M:%S.%f%Z')
        except ValueError:
            d = datetime.datetime.strptime(timestamp, '%a, %d %b %Y %H:%M:%S %Z')
        return d

    def upload_large_file(self, path):
        log.info('Uploading {} to S3 storage (large file)'.format(path))
        file_name = os.path.basename(path)
        multipart_file_name = file_name + "multipart"
        file_size = os.stat(path).st_size
        mp = self.s3bucket.initiate_multipart_upload(self.expand_path(multipart_file_name))
        segment_id = 0
        try:
            with open(path, "rb") as obj:
                while self.segment_size * segment_id < file_size:
                    segment_id += 1
                    log.info("Uploading segment {}".format(segment_id))
                    mp.upload_part_from_file(fp=obj, part_num=segment_id, size=self.segment_size)
                mp.complete_upload()
        except IOError:
            log.info("Unable to read {}".format(path))
            return Status.FAILURE
        uploaded = self.s3bucket.get_key(self.expand_path(multipart_file_name))
        if uploaded:
            # Copy the file (within S3) to get a "normal" etag
            copy = uploaded.copy(self.s3bucket, self.expand_path(file_name))
            uploaded.delete()
            if not copy:
                log.info("Post-upload copy failed for {}".format(path))
                return Status.FAILURE
            return Status.SUCCESS
        log.info("Upload failed for {}".format(path))
        return Status.FAILURE

    def upload_file(self, path):
        """ A direct push of the file at path to cloud storage, no md5 compare just overwrite
        :param path: The local path of the file to upload
        """
        file_size = os.stat(path).st_size
        if file_size > self.segment_size:
            return self.upload_large_file(path)
        log.info('Uploading {} to S3 storage'.format(path))

        file_name = os.path.basename(path)
        s3object = boto.s3.key.Key(self.s3bucket)
        s3object.key = self.expand_path(file_name)
        try:
            bytes_written = s3object.set_contents_from_filename(path)
        except IOError:
            log.info("Unable to read {}".format(path))
            return Status.FAILURE

        if bytes_written != file_size:
            log.info("Upload failed! {} size: {} written: {}".format(path, file_size, bytes_written))
            return Status.FAILURE
        else:
            return Status.SUCCESS

    def delete(self, name):
        """ Delete an object from cloud storage
        :param name: The name of the object to delete
        :return:
        """
        log.info('Deleting object {} from bucket {}'.format(self.expand_path(name), self.bucket))
        s3object = self.s3bucket.get_key(self.expand_path(name))
        s3object.delete()
        return Status.SUCCESS

    def info(self, name=None):
        """ Get information on an object
        Returned object information must contain the following fields:
        name, size, contentType, updated, md5Hash
        :param name: If name is None, provide bucket information otherwise object information
        :return: A dictionary representation of the bucket or object information.
        """
        if not name:
            # Bucket information
            # TODO
            return {}

        # Object information
        s3object = self.s3bucket.get_key(self.expand_path(name))
        details = {}
        if s3object:
            details['name'] = name
            details['size'] = s3object.size
            details['contentType'] = s3object.content_type
            details['updated'] = s3object.last_modified
            details['md5Hash'] = s3object.etag.replace('"', '')
        return details

    def list_objects(self, name=None):
        """ List objects in self.bucket, this lists all objects even if multiple requests to the
            api are required to retreive them all.
            Note: Swift doesn't store creation time, so this will be the same as updated time
        :param name: If name is none list all objects in the container
        :return: A list of dictionaries with these fields: name, size, contentType, timeCreated, updated, md5Hash
                  the list is empty if nothing is found.
        """
        objects = []
        for obj in self.s3bucket.list(prefix=self.expand_path(name)):
            details = {}
            details["name"] = obj.name
            details["size"] = obj.size
            details["contentType"] = obj.content_type
            details["timeCreated"] = obj.last_modified
            details["updated"] = obj.last_modified
            details["md5Hash"] = obj.etag.replace('"', '')
            objects.append(details)
        return objects


if __name__ == "__main__":
    sys.exit(main(S3CloudStorage))
