#!/usr/bin/python3
#
# -------------------------------------------------------#
#  This file is Juju managed - do not make local changes #
# -------------------------------------------------------#
#
# Author: Chris Stratford <chris.stratford@canonical.com>
# Copyright: Canonical Ltd. 2014
#
# Archive log files to swift storage
#
# Should be quiet on success, otherwise reports any errors to stdout
# Useful info logged to a status file (file overwritten each run)
#
# The standard invocation setup by the charm is
# 'archive-to-swift.py -s -f [config_file] -l [logdir1] -l [logdir2] -c [container_name] -a [archive_after] -d [delete_after] -p [retention_days]'

import os
import logging
import sys
import datetime

import json
import swiftclient
from archive import CloudStorage, main, Status

log = logging.getLogger()


class SwiftCloudStorage(CloudStorage):
    """ A class wrapping the interface to Swift storage
        The intent is to hide the complexity of login and other common operations.
        bucket = container
        project = storage account
    """
    def __init__(self, json_file, bucket, project=None):
        creds = json.loads(open(json_file, 'r').read())
        wanted = ["os_username",
                  "os_password",
                  "os_tenant_name",
                  "os_auth_url",
                  "os_region_name"]
        for k in wanted:
            if k not in creds:
                # Missing creds
                log.info("Missing credentials ({}) in config file".format(k))
                raise KeyError
        options = {}
        options["tenant_name"] = creds["os_tenant_name"]
        options["region_name"] = creds["os_region_name"]
        if "os_storage_url" in creds:
            options["object_storage_url"] = creds["os_storage_url"]

        self.swift = swiftclient.client.Connection(
            authurl=creds["os_auth_url"],
            user=creds["os_username"],
            key=creds["os_password"],
            retries=5,
            auth_version="2.0",
            os_options=options)

        self.segment_size = 1073741824
        self.bucket, self.bucket_path = bucket.split('/', 1)
        self.project = project

        # Ensure the bucket exists before using it
        try:
            self.swift.head_container(self.bucket)
        except swiftclient.ClientException:
            self.swift.put_container(self.bucket)

    def _calculate_md5(self, path, style="hexdigest"):
        """ Calculate and return an md5 for a local file.
        :param path: The local filesystem path to calculate an md5 for
        :param style: The output format, one of "base64" or "hexdigest" (default)
        :return: The md5 hash
        """
        return super(SwiftCloudStorage, self)._calculate_md5(path, style=style)

    def standard_timestamp(self, timestamp):
        """ Convert a GCE timestamp to something standardised
        Swift returns a timestamp like "2016-03-11T15:26:05.814010" when listing objects
        and "Fri, 11 March 2016 15:26:05 GMT" when stat-ing them, so we handle both
        :param timestamp: A file timestamp in one of Swift's formats
        :return: timestamp in standardised format
        """
        try:
            d = datetime.datetime.strptime(timestamp, '%Y-%m-%dT%H:%M:%S.%f')
        except ValueError:
            d = datetime.datetime.strptime(timestamp, '%a, %d %b %Y %H:%M:%S %Z')
        return d

    def upload_large_file(self, path):
        log.info('Uploading {} to Swift storage (large file)'.format(path))
        file_name = os.path.basename(path)
        file_size = os.stat(path).st_size

        try:
            md5 = self._calculate_md5(path)
        except IOError:
            log.info("Unable to read {}".format(path))
            return Status.FAILURE

        # Ensure the segment bucket exists
        segment_bucket = '{}_segments'.format(self.bucket)
        try:
            self.swift.head_container(segment_bucket)
        except swiftclient.ClientException:
            self.swift.put_container(segment_bucket)

        segment_id = 0
        mtime = str(os.path.getmtime(path))
        # We know it's readable as we've done the md5sum, so no need for try/except here
        with open(path, "rb") as obj:
            while self.segment_size * segment_id < file_size:
                data = obj.read(self.segment_size)
                segment_name = '{}/{}/{}/{}/{:08d}'.format(file_name, mtime, file_size, self.segment_size, segment_id)
                segment_id += 1
                log.info("Uploading segment {}".format(segment_id))
                try:
                    self.swift.put_object(segment_bucket,
                                          self.expand_path(segment_name),
                                          data,
                                          content_type="application/octet-stream")
                except swiftclient.ClientException:
                    log.info("Segment upload failed!")
                    return Status.FAILURE

        # New we have all the part uploaded, create the manifest
        manifest = '{}/{}/{}/{}/{}/'.format(segment_bucket, self.expand_path(file_name), mtime, file_size, self.segment_size)
        headers = {
            'x-object-manifest': manifest,
            'x-object-meta-mtime': mtime,
            'x-object-meta-file-md5': md5,
        }
        try:
            self.swift.put_object(self.bucket,
                                  self.expand_path(file_name),
                                  '',
                                  content_length=0,
                                  content_type="application/octet-stream",
                                  headers=headers)
        except swiftclient.ClientException:
            log.info("Manifest upload failed!")
            return Status.FAILURE

        return Status.SUCCESS

    def upload_file(self, path):
        """ A direct push of the file at path to cloud storage, no md5 compare just overwrite
        :param path: The local path of the file to upload
        """
        file_size = os.stat(path).st_size
        if file_size > self.segment_size:
            return self.upload_large_file(path)
        log.info('Uploading {} to Swift storage'.format(path))

        file_name = os.path.basename(path)

        try:
            with open(path, "rb") as obj:
                try:
                    self.swift.put_object(self.bucket,
                                          self.expand_path(file_name),
                                          obj,
                                          content_type="application/octet-stream")
                except swiftclient.ClientException:
                    log.info("Upload failed")
                    return Status.FAILURE
        except IOError:
            log.info("Unable to read {}".format(path))
            return Status.FAILURE
        return Status.SUCCESS

    def delete(self, name):
        """ Delete an object from cloud storage
        :param name: The name of the object to delete
        :return:
        """
        log.info('Deleting object {} from bucket {}'.format(self.expand_path(name), self.bucket))
        self.swift.delete_object(self.bucket, name)
        return Status.SUCCESS

    def info(self, name=None):
        """ Get information on an object
        Returned object information must contain the following fields:
        name, size, contentType, updated, md5Hash
        :param name: If name is None, provide bucket information otherwise object information
        :return: A dictionary representation of the bucket or object information.
        """
        if not name:
            # Bucket information
            metadata = self.swift.head_container(self.bucket)
            details = {}
            details["name"] = name,
            details["size"] = int(metadata["x-container-bytes-used"])
            details["contentType"] = metadata["content-type"]
            details["timeCreated"] = metadata["x-timestamp"]
            details["updated"] = metadata["x-timestamp"]
            details["md5Hash"] = None
            return details

        # Object details
        metadata = self.swift.head_object(self.bucket, self.expand_path(name))
        details = {}
        details["name"] = name
        # Zero-length objects don't get content-length when proxied
        try:
            details["size"] = int(metadata["content-length"])
        except KeyError:
            details["size"] = 0
        details["contentType"] = metadata["content-type"]
        details["timeCreated"] = metadata["last-modified"]
        details["updated"] = metadata["last-modified"]
        # Prefer our own md5 header if available (handling for large files)
        if 'x-object-meta-file-md5' in metadata:
            details['md5Hash'] = metadata['x-object-meta-file-md5']
        else:
            details['md5Hash'] = metadata['etag'].replace('"', '')
        return details

    def list_objects(self, name=None):
        """ List objects in self.bucket/self.bucket_path, this lists all objects even if multiple
            requests to the api are required to retreive them all.
            Note: Swift doesn't store creation time, so this will be the same as updated time
        :param name: If name is none list all objects in the container
        :return: A list of dictionaries with these fields: name, size, contentType, timeCreated, updated, md5Hash
                  the list is empty if nothing is found.
        """
        objects = []
        metadata, data = self.swift.get_container(self.bucket, prefix=self.expand_path(name))
        for obj in data:
            details = {}
            details["name"] = obj["name"]
            details["size"] = int(obj["bytes"])
            details["contentType"] = obj["content_type"]
            details["timeCreated"] = obj["last_modified"]
            details["updated"] = obj["last_modified"]
            details["md5Hash"] = obj["hash"]
            objects.append(details)
        return objects


if __name__ == "__main__":
    sys.exit(main(SwiftCloudStorage))
