#!/usr/bin/python3
#
# -------------------------------------------------------#
#  This file is Juju managed - do not make local changes #
# -------------------------------------------------------#
#
# Author: Tim Kuhlman <timothy.kuhlman@canonical.com>
# Copyright: Canonical Ltd. 2015
#
# This file contains the common code used by provider specific archive_to_* scripts.

import argparse
import base64
import datetime
import glob
import hashlib
import logging
import os
import subprocess
import tempfile
import time

from preprocessor import get_preprocessor

log = logging.getLogger()


class Status:
    FAILURE = 0
    SUCCESS = 1
    SKIPPED = 2


class CloudStorage(object):
    """ Abstract class to define an interface common to many storage backends.
    """
    def __init__(self, config_file, container):
        """
        :param config_file: A config file with relavant credentials for logging in
        :param container: The container to operate with
        """
        raise NotImplementedError

    # Much of this method was taken from the the original archive-to-* scripts and updated for pep8
    # Memory efficient checksum code reading in blocks of data.
    # http://stackoverflow.com/questions/3431825/generating-a-md5-checksum-of-a-file
    @staticmethod
    def _calculate_md5(path, style="base64"):
        """ Calculate and return a base64 md5 for a local file.
        :param path: The local filesystem path to calculate an md5 for
        :param style: The output format, one of "base64" (default) or "hexdigest"
        :return: The md5 hash
        """
        afile = open(path, 'rb')
        hasher = hashlib.md5()

        # First find out how large the file is.
        fsize = os.fstat(afile.fileno()).st_size

        bytes_read = 0
        while True:
            buf = afile.read(65536)  # the read size could be made configurable
            bytes_read += len(buf)
            hasher.update(buf)

            if bytes_read >= fsize:
                break

        if style == "hexdigest":
            return hasher.hexdigest()
        else:
            return base64.b64encode(hasher.digest())

    def _calculate_str_md5(self, s, style="base64"):
        """ Calculate and return a base64 md5 for a string.
        :param s: A string to calculate the md5 for
        :param style: The output format, one of "base64" (default) or "hexdigest"
        :return: The md5 hash
        """
        hasher = hashlib.md5()
        hasher.update(s)
        if style == "hexdigest":
            return hasher.hexdigest()
        else:
            return base64.b64encode(hasher.digest())

    def standard_timestamp(self, timestamp):
        """ Convert a cloud-specific timestamp into something standard
        :param timestamp: A file timestamp in GCE format
        :return: timestamp in standardised format
        """
        raise NotImplementedError

    def delete(self, name):
        """ Delete an object from cloud storage
        :param name: The name of the object to delete
        :return:
        """
        raise NotImplementedError

    def expand_path(self, name=None):
        """ Expand an object name to include its bucket_path
        :param name: The object name (e.g. logfile.log)
        :return: A combined path and object name (e.g. unit-1/logfile.log)
        If no name is provided, just the bucket_path will be returned
        If name is already expanded, it will be returned as-is
        """
        if not name:
            return self.bucket_path + "/"
        if name.startswith(self.bucket_path + "/"):
            return name
        return self.bucket_path + "/" + name

    def strip_path(self, name=None):
        """ Strip the bucket_path from an obect name
        :param name: The object name (e.g. unit-1/logfile.log)
        :return: A combined path and object name (e.g. logfile.log)
        If no name is provded, it will return an empty string
        If name is already stripped, it will be returned as-is
        """
        return name.rpartition(self.bucket_path + "/")[2]

    def upload_file(self, path):
        """ A direct push of the file at path to the cloud storage, no md5 compare just overwrite
        :param path: The local path of the file to upload
        """
        raise NotImplementedError

    def info(self, name=None):
        """
        Returned object information must contain the following fields:
        name, size, contentType, updated, md5Hash
        :param name: If name is None, provide bucket information otherwise object information
        :return: A dictionary representation of the bucket or object information.
        """
        raise NotImplementedError

    def list_objects(self, name=None):
        """ Return a list of dictionaries of object details, for objects the container.
        :param name: If name is none list all objects in the container
        :return: A list of dictionaries with these fields name, size, contentType, timeCreated, updated, md5Hash.
                  the list is empty is nothing is found.
        """
        raise NotImplementedError

    def get_archived_files(self):
        self.archived_files = {}
        for obj in self.list_objects():
            self.archived_files[self.strip_path(obj["name"])] = obj

    def needs_archiving(self, path):
        """ Decide if we need to archive this file or not.
            Returns True if it does (file differs from uploaded one or is new)
            or Flase if it doesn't (file is the same as uploaded one)
        """

        if os.path.basename(path) in self.archived_files:
            return False

        try:
            info = self.info(os.path.basename(path))
        except Exception:
            # TODO(pjdc): We should throw appropriate exceptions from the
            # storage drivers but for now this will have to do.
            return True

        try:
            local_size = os.stat(path).st_size
        except FileNotFoundError:
            # This can happen in the case of logs that that hasn't been
            # pre-processed yet.
            return True

        if info['size'] == local_size:
            # Sizes match, how about md5sums?
            local_md5 = self._calculate_md5(path)
            if info['md5Hash'] == local_md5:
                # MD5 matches, no need to re-upload
                return False

        return True


def configure_logging(args):
    """ Configure logging to go to a status file at INFO level and output
        as deteremined by arguments. The status file name is derived from
        the configuration file path/name and is ovewritten on each run
    """
    status_file_name = "cla-{}.status".format(os.path.splitext(os.path.basename(args.config_file))[0])
    status_file = os.path.join(os.path.dirname(args.config_file), status_file_name)
    log.setLevel(logging.DEBUG)  # The root logger level
    if os.path.exists(status_file):
        os.remove(status_file)
    fh = logging.FileHandler(status_file)
    fh.setLevel(logging.INFO)
    fh.setFormatter(logging.Formatter('%(message)s'))
    log.addHandler(fh)

    # Console logging defaults to warn but can be modified
    ch = logging.StreamHandler()
    if args.debug:
        ch.setLevel(logging.DEBUG)
    elif args.verbose:
        ch.setLevel(logging.INFO)
    else:
        ch.setLevel(logging.WARN)
    ch.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
    log.addHandler(ch)


def expand_globs(dirs):
    expanded = []
    for d in dirs:
        expanded += glob.glob(d)
    return expanded


def get_files_to_archive(dirs, archive_after):
    """
    :param dirs: A list of log dirs, files or glob patterns
    :param archive_after: Only archive files that are at least this many days old
    :return: A list of paths to files that should be archived
    """
    log.info('Parsing log directories to find files to archive')
    cutoff = time.time() - (86400 * int(archive_after))
    log_files = []
    for log_dir in expand_globs(dirs):
        if os.path.isfile(log_dir):
            log_files.append(log_dir)
            continue
        if not os.path.isdir(log_dir):
            log.warn("Log dir '{}' not found - skipping".format(log_dir))
            continue

        for root, dirs, files in os.walk(log_dir):
            for f in files:
                path = os.path.join(root, f)
                if os.path.getmtime(path) < cutoff:
                    log_files.append(path)

    log_files.sort()
    return log_files


def parse_args():
    parser = argparse.ArgumentParser(description="Upload log files to Google Compute Engine Cloud storage")
    parser.add_argument("-l", dest="log_dirs", action="append", required=True,
                        help="Log directory to be archived (can be repeated)")
    parser.add_argument("-c", dest="container", required=True,
                        help="Bucket/Container name and path")
    parser.add_argument("-a", dest="archive_after", required=True, type=int,
                        help="How old (in days) a file should be before archiving")
    parser.add_argument("-d", dest="delete", action="store_true",
                        help="Delete the local file if successfully archived")
    parser.add_argument("--debug", dest="debug", action="store_true",
                        help="Debugging output")
    parser.add_argument("-s", dest="sudo", action="store_true",
                        help="Use 'sudo rm' when deleting files (default: rm)")
    parser.add_argument("--pre", dest="preprocessors", action="append",
                        help="Preprocessors to run on the logs files before uploading")
    parser.add_argument("-p", dest="purge_after", type=int,
                        help="How old (in days) to keep files in swift before purging")
    parser.add_argument("-v", dest="verbose", action="store_true",
                        help="Be verbose")
    parser.add_argument("-f", dest="config_file",
                        help="JSON key config file for the service account you wish to use.")
    parser.add_argument("--project", dest="project",
                        help="The Google project to create the container under. Only needed for container creation.")
    return parser.parse_args()


def preprocess_log_filename(file_path, preprocessor_list):
    """ Report what the final filename should be after preprocessing
    :param file_path: The path to the original logfile.
    :param preprocessor_list: A list of strings representing the preprocessor name and arguments
    :return: The path the processed file should have
    """
    proposed_file_path = file_path
    for processor_str in preprocessor_list:
        splits = processor_str.split(' ', 1)
        pname = splits[0]
        # Ensure backwards compatibility for the obsolete name of the anonymizer preprocessor.
        if pname == 'anonymize_web_log_ips':
            pname = 'anonymize_web_log'

        processor = get_preprocessor(pname)
        if processor is None:
            log.error('No preprocessor found for {}.'.format(pname))
            continue
        proposed_file_path = processor.filename(proposed_file_path)

    return proposed_file_path


def preprocess_log(file_path, preprocessor_list):
    """ Run any pre-processors on the log file writing out a new temporary log file which is returned
    :param file_path: The path to the original logfile.
    :param preprocessor_list: A list of strings representing the preprocessor name and arguments
    :return: The path to the processed file
    """
    proposed_file_path = file_path
    work_dir = os.path.dirname(file_path)
    processed_file = tempfile.NamedTemporaryFile(delete=False, dir=work_dir)
    with open(file_path, 'rb') as log_file:
        for processor_str in preprocessor_list:
            splits = processor_str.split(' ', 1)
            pname = splits[0]
            if len(splits) > 1:
                args = splits[1]
            else:
                args = None
            processor = get_preprocessor(pname)
            if processor is None:
                log.error('No preprocessor found for {}.'.format(pname))
                continue
            # I abuse the log_file variable here to string together a bunch of preprocessors or none as the case may be
            old_log_file = log_file
            log_file = processor.process(old_log_file, args)
            proposed_file_path = processor.filename(proposed_file_path)
            if log_file != old_log_file:
                old_log_file.close()

        processed_file.write(log_file.read())
    os.rename(processed_file.name, proposed_file_path)

    return proposed_file_path


def main(cloud_storage):
    args = parse_args()  # sys.argv is assumed
    configure_logging(args)

    log.info("DATE: {}".format(time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())))
    log.info('Logging into the cloud provider')
    store = cloud_storage(args.config_file, args.container, args.project)

    archived_count = 0
    failed_count = 0

    # Get a list of files already archived
    store.get_archived_files()

    log.info("Preprocessors: {}".format(args.preprocessors))
    for log_file in get_files_to_archive(args.log_dirs, args.archive_after):
        # PreProcess
        if args.preprocessors is not None:
            processed_file = preprocess_log_filename(log_file, args.preprocessors)
        else:
            processed_file = log_file

        log.info("processed_file: {}".format(processed_file))
        if store.needs_archiving(processed_file):
            if args.preprocessors is not None:
                processed_file = preprocess_log(log_file, args.preprocessors)
            # Upload to storage provider (and tidyup if successful)
            if store.upload_file(processed_file) == Status.SUCCESS:
                archived_count += 1
                if processed_file != log_file:
                    os.remove(processed_file)

                if args.delete:
                    log.info('Remove file {}'.format(log_file))
                    if args.sudo:
                        rm_cmd = ['sudo', '-n', 'rm', log_file]
                    else:
                        rm_cmd = ['rm', log_file]
                    try:
                        subprocess.check_output(rm_cmd, shell=False)
                    except subprocess.CalledProcessError as err:
                        log.error("Error running: {}\n{}".format(rm_cmd, err))
            else:
                failed_count += 1

    log.info("Archived: {} of {} files".format(archived_count, archived_count + failed_count))
    if args.purge_after and failed_count == 0:  # Skip purge if failures occur
        log.info('Purging files from cloud storage older than {} days'.format(args.purge_after))
        purge_delta = datetime.timedelta(days=args.purge_after)
        deleted_count = 0
        for log_file in store.list_objects():
            updated_at = store.standard_timestamp(log_file['updated'])
            if updated_at < datetime.datetime.now() - purge_delta:
                store.delete(log_file['name'])
                deleted_count += 1
        log.info("Deleted: {} files".format(deleted_count))
