#!/usr/bin/python3
#
# -------------------------------------------------------#
#  This file is Juju managed - do not make local changes #
# -------------------------------------------------------#
#
# Author: Tim Kuhlman <timothy.kuhlman@canonical.com>
# Copyright: Canonical Ltd. 2015
#
# This file contains the code for the pre-processors

import inspect
import gzip
import os.path
import sys
import tempfile

import anonymize_web_log

BLOCKSIZE = 1024 * 1024 * 100    # Read files in 100MB chunks


class PreProcessor(object):
    """ Abstract class to define a common interface for all preprocessors
    """
    name = 'ABC'  # Defines the name this class will be identified by for get_preprocessors

    def filename(self, file_name):
        """
        :param file_name: The current file name
        :return The suggested post-processing file name
        """
        raise NotImplementedError

    def process(self, log_file, args):
        """
        :param log_file: A file object to read from
        :param args: Any arguments for defining the behavior of the processing
        :return A file object that has been written to
        """
        raise NotImplementedError


class AnonymizeWebLog(PreProcessor):
    """ Runs the anonymize_web_log.py script as a preprocessor
    """
    name = 'anonymize_web_log'

    def filename(self, file_name):
        return file_name + ".anon"

    def process(self, log_file, args):
        """
        :param log_file: A file object to read from
        :param args: Any arguments for defining the behavior of the processing
        :return A file object that has been written to
        """
        new_file = tempfile.SpooledTemporaryFile(max_size=1024)
        # Setup fake_args to get past the parsing, then don't use the file/output args like standalone anonymize
        if args is None:
            fake_args = ['none']
        else:
            fake_args = args.split()
            fake_args.append('none')

        parsed_args = anonymize_web_log.parse_args(fake_args)
        anonymize_web_log.anonymize(new_file,
                                    log_file,
                                    parsed_args.skip_ips,
                                    parsed_args.skip_private,
                                    parsed_args.strip_usernames,
                                    parsed_args.username_field)
        new_file.seek(0)

        return new_file


class Gzip(PreProcessor):
    """ A gzip preprocessor
    """
    name = 'gzip'

    def filename(self, file_name):
        return file_name + ".gz"

    def process(self, log_file, args):
        """
        :param log_file: A file object to read from
        :param args: Any arguments for defining the behavior of the processing
        :return A file object that has been written to
        """
        new_file = tempfile.SpooledTemporaryFile(max_size=1024)
        gfile = gzip.GzipFile(filename='', mode='wb', fileobj=new_file)
        while True:
            try:
                data = log_file.read(BLOCKSIZE)
            except IOError:
                # read error - return the original file
                gfile.close()
                new_file.close()
                new_file = log_file
                break
            if not data:
                gfile.close()
                break
            try:
                gfile.write(data)
            except IOError:
                # Write error on gzipped file - return the original
                gfile.close()
                new_file.close()
                new_file = log_file
                break
        new_file.seek(0)

        return new_file


class Gunzip(PreProcessor):
    """ A preprocessor that gunzips the file
    """
    name = 'gunzip'

    def filename(self, file_name):
        if file_name.endswith(".gz"):
            return os.path.splitext(file_name)[0]
        else:
            return file_name

    def process(self, log_file, args):
        """
        :param log_file: A file object to read from
        :param args: Any arguments for defining the behavior of the processing
        :return A file object that has been written to
        """
        new_file = tempfile.SpooledTemporaryFile(max_size=1024)
        gfile = gzip.GzipFile(filename='', mode='rb', fileobj=log_file)
        while True:
            try:
                data = gfile.read(BLOCKSIZE)
            except IOError:
                # Can't uncompress (may not be gzipped) - return the original file
                new_file.close()
                new_file = log_file
                break
            if not data:
                break
            try:
                new_file.write(data)
            except IOError:
                # Write error on temporary file - return the original
                new_file.close()
                new_file = log_file
                break
        gfile.close()
        new_file.seek(0)

        return new_file


def get_preprocessor(name):
    """ Finds a subclass of PreProcessor that matches name
    :param name: The string to match to the name attribute of the class
    :return: An object of the PreProcessor subclass that matches name or None
    """
    # Only looking in this file for now but if preprocessors become popular we could expand it to others
    classes = inspect.getmembers(sys.modules[__name__], inspect.isclass)
    preprocessor_class = None
    for _, clsmember in classes:
        if issubclass(clsmember, PreProcessor) and clsmember.name == name:
            preprocessor_class = clsmember
            break

    if preprocessor_class is None:
        return None
    else:
        return preprocessor_class()
