#!/usr/bin/python3
#
# Perform an MD5 check of a local file against swift
# Should work with both single and split swift files

import os
import sys
from swiftclient import client
import hashlib


class Swiftmd5:
    # Connect to swift
    def connect(self):
        userName = os.environ['OS_USERNAME']
        userPass = os.environ['OS_PASSWORD']
        userTenant = os.environ['OS_TENANT_NAME']
        authURL = os.environ['OS_AUTH_URL']
        regionName = os.environ['OS_REGION_NAME']

        swift = client.Connection(authURL, userName, userPass, 5, auth_version='2.0', os_options={'tenant_name': userTenant, 'region_name': regionName})
        return swift

    # MD5 a file (possibly in several chunks)
    def md5sums(self, swiftChunkSize, uploadFile):
        readChunk = 102400
        md5sums = []

        m = hashlib.md5()
        count = 0
        with open(uploadFile, "r") as f:
            while True:
                line = f.read(readChunk)
                if not line:
                    if count > 0:
                        md5sums.append(m.hexdigest())
                    break
                count += readChunk
                m.update(line)
                if count == swiftChunkSize:
                    md5sums.append(m.hexdigest())
                    m = hashlib.md5()
                    count = 0
        return md5sums

    # Compare the local and swift files
    def compare(self, container, uploadFile, swiftChunkSize):
        swift = self.connect()
        metadata = swift.head_object(container, uploadFile)

        if "x-object-manifest" in metadata:
            # Split file
            splitContainer, sep, splitPath = metadata["x-object-manifest"].partition("/")
            splitMetadata, splitData = swift.get_container(splitContainer, prefix=splitPath)
            hashes = [None] * len(splitData)
            for segment in splitData:
                num = int(segment["name"].rpartition("/")[2])
                hashes[num] = segment["hash"]
            filemd5 = self.md5sums(swiftChunkSize, uploadFile)
            if len(hashes) != len(filemd5):
                return False
            for i in range(0, len(hashes)):
                if filemd5[i] != hashes[i]:
                    return False
            return True
        else:
            # Single file
            filemd5 = self.md5sums(-1, uploadFile)[0]
            swiftmd5 = metadata["etag"]
            if filemd5 == swiftmd5:
                return True
            return False


if __name__ == '__main__':
    container = sys.argv[1]
    uploadFile = sys.argv[2]
    swiftChunkSize = int(sys.argv[3])

    if not (swiftChunkSize and uploadFile and container):
        print("Usage: {} <container> <file> <chunk size>".format(os.path.basename(sys.argv[0])))
        sys.exit(1)

    s = Swiftmd5()
    if s.compare(container, uploadFile, swiftChunkSize):
        print("MD5 matches")
        sys.exit(0)
    else:
        print("MD5 comparison failed")
        sys.exit(1)
