#!/usr/bin/python3
#
# Copyright 2014 Canonical Ltd.  All rights reserved
# Author: Chris Stratford <chris.stratford@canonical.com>

import pwd
import base64
import json
import os
import sys
import socket
sys.path.insert(0, os.path.join(os.environ['CHARM_DIR'], 'lib'))
import charmhelpers.core.hookenv as hookenv


class Config:
    _relation_keys = ["logdirs", "archive_after", "parent_unit_name",
                      "nagios_hostname", "nagios_host_context",
                      "retention_days", "private-address"]
    _relation_data = {}

    def __init__(self):
        self.config = hookenv.config()

    def config_file(self):
        config_dir = self.archive_userinfo().pw_dir
        return os.path.join(config_dir, "log-archive.conf")

    def anonymize_web_log_options(self):
        return str(self.config.get("anonymize_web_log_options"))

    def archive_after(self):
        try:
            archive_after = self.relation_data()["archive_after"]
        except KeyError:
            archive_after = str(self.config.get("archive_after_default"))
        return archive_after

    def archive_userinfo(self):
        try:
            archive_userinfo = pwd.getpwnam(self.archive_username())
        except KeyError:
            hookenv.log("CHARM: archive user '{}' has not been created".format(self.archive_username()))
            raise KeyError
        return archive_userinfo

    def archive_username(self):
        return str(self.config.get("username"))

    def container_credentials(self):
        return base64.b64decode(self.config.get("container_credentials")).decode('utf-8')

    def container_name(self):
        return str(self.config.get("container_name"))

    def container_type(self):
        return str(self.config.get("container_type"))

    def container_type_changed(self):
        return self.config.changed("container_type")

    def cron_time(self):
        return str(self.config.get("cron_time"))

    def delete_after_archive(self):
        return self.config.get("delete_after_archive")

    def project(self):
        return self.config.get("project")

    def retention_days(self):
        try:
            value = self.relation_data()["retention_days"]
        except KeyError:
            return None
        return value

    def logdirs(self):
        fromConfig = []
        for logdir in str(self.config.get("logdirs")).split():
            fromConfig.append(logdir)
        try:
            fromRelation = self.relation_data()["logdirs"]
        except KeyError:
            fromRelation = []
        return fromRelation + fromConfig

    def nagios_host_context(self):
        try:
            host_context = self.relation_data()["nagios_host_context"]
        except KeyError:
            host_context = str(self.config.get("nagios_host_context"))
        return host_context

    def nagios_hostname(self):
        try:
            hostname = self.relation_data()["nagios_hostname"]
        except KeyError:
            hostname = socket.gethostname()
        return hostname

    def nagios_relation_ok(self):
        try:
            if self.nagios_hostname():
                return True
            else:
                return False
        except KeyError:
            return False

    def log_archive_rel_ok(self):
        # Do we have a log_archive relation?
        if set(("logdirs", "archive_after")).issubset(self.relation_data()):
            hookenv.log("Found container-log-archive relation data")
            return True
        # Maybe we have a juju-info relation and some locally set logdirs?
        if "private-address" in self.relation_data() and self.logdirs():
            hookenv.log("Found locally defined logdirs")
            return True
        hookenv.log("No relation or local logdirs available")
        return False

    def nagios_servicegroup(self):
        servicegroup = str(self.config.get("nagios_servicegroup"))
        if not servicegroup:
            servicegroup = "{}-{}".format(self.nagios_host_context(), hookenv.service_name())
        return servicegroup

    def parent_unit_name(self):
        return self.relation_data()["parent_unit_name"]

    def preprocessors(self):
        return str(self.config.get("preprocessors")).split()

    def relation_data(self):
        if not hookenv.in_relation_hook():
            if self._relation_data:
                return self._relation_data

        # Load any saved data
        try:
            with open(self.config_file(), "r") as f:
                if os.path.getsize(self.config_file()) > 0:
                    self._relation_data = json.loads(f.read())
        except IOError:
            # File may not be there initially - this is ok
            pass

        if not hookenv.in_relation_hook():
            return self._relation_data

        hookenv.log("CHARM: In relation hook {}".format(hookenv.relation_type()))
        for key in self._relation_keys:
            val = hookenv.relation_get(key)
            if val:
                if key == "logdirs":
                    # Expand the list
                    logdirs = []
                    for logdir in val.split():
                        logdirs.append(str(base64.b64decode(logdir)).rstrip())
                    val = logdirs
                self._relation_data[key] = val

        # Which unit is my primary relation with?
        if "remote_unit" not in self._relation_data:
            if "log-archive" in hookenv.relations():
                candidates = hookenv.relations()["log-archive"]
            elif "juju-info" in hookenv.relations():
                candidates = hookenv.relations()["juju-info"]
            else:
                candidates = {}
            for rel_id in candidates:
                for unit in candidates[rel_id]:
                    if unit != hookenv.local_unit():
                        hookenv.log("CHARM: likely primary: {}".format(unit))
                        self._relation_data["remote_unit"] = unit

        self._relation_data["_This file is auto-generated.  Do not edit_"] = True
        with open(self.config_file(), "w") as f:
            f.write(json.dumps(self._relation_data, sort_keys=True, indent=4))
        return self._relation_data

    def container_path(self):
        if "remote_unit" in self.relation_data():
            value = self.relation_data()["remote_unit"].replace('/', '-')
            hookenv.log("CHARM: Choosing container path, {} from remote_unit".format(value))
        else:
            value = socket.gethostname().replace('/', '-')
            hookenv.log("CHARM: No container path found - using hostname")
        return value

    def apt_repository(self):
        return str(self.config.get("apt_repository"))

    def apt_repository_key(self):
        return str(self.config.get("apt_repository_key"))
