#!/usr/bin/python3
#
# Copyright 2014 Canonical Ltd.  All rights reserved
# Author: Chris Stratford <chris.stratford@canonical.com>

import sys
import random
import os
import stat
import os.path
import socket
import time
sys.path.insert(0, os.path.join(os.environ['CHARM_DIR'], 'lib'))
from charmhelpers.core.host import mkdir, adduser, service_reload
from charmhelpers.core.hookenv import Hooks, log, charm_dir, service_name
from charmhelpers.fetch import apt_install, apt_update, add_source
from jinja2 import Template
from Config import Config

hooks = Hooks()
hostname = socket.gethostname()
conf = Config()

required_pkgs = {
    "common": [
        "gzip",
        "python3-netaddr",
        "python3-httplib2",
        "xz-utils",
    ],
    "azure": [],
    "gce": [
        "python3-googleapi",
    ],
    "swift": [
        "python3-keystoneclient",
        "python3-swiftclient",
    ],
    "s3": [
        "python3-boto",
    ],
}


def juju_header():
    header = ("#-------------------------------------------------#\n"
              "# This file is Juju managed - do not edit by hand #\n"
              "#-------------------------------------------------#\n")
    return header


def file_from_template(tmpl, dest, searchList):
    template_file = os.path.join(charm_dir(), "templates", tmpl)
    with open(template_file, 'r', encoding='utf-8') as t_file:
        t = Template(t_file.read())
    with open(dest, "w") as f:
        f.write(juju_header())
        f.write(t.render(searchList))
    os.chmod(dest, 0o444)


def expand_crontimes(crontimes):
    cron_ranges = [(0, 59), (0, 23), (1, 31), (1, 12), (0, 6)]
    timelist = crontimes.split()
    for i in range(len(timelist)):
        if timelist[i] == "<<rnd>>":
            timelist[i] = str(random.randint(*cron_ranges[i]))

    expanded_times = " ".join(timelist)
    return expanded_times


def install_packages(container_type):
    try:
        aptsrc = conf.apt_repository()
        try:
            aptkey = conf.apt_repository_key()
        except KeyError:
            aptkey = None
    except KeyError:
        # No apt source, continue without it
        pass
    else:
        add_source(aptsrc, aptkey)

    apt_update()
    apt_install(required_pkgs["common"], options=['--force-yes'])
    apt_install(required_pkgs[container_type], options=['--force-yes'])


def parent_is_ready():
    try:
        conf.parent_unit_name()
    except KeyError:
        log("CHARM: Parent unit not ready yet")
        return False
    return True


# Remove any old service__*_check_log_archive.cfg files (in case of hostname change)
def wipe_nrpe_exports(nagios_exportdir):
    if not os.path.isdir(nagios_exportdir):
        return
    for f in os.listdir(nagios_exportdir):
        if f.startswith("service__") and f.endswith("_check_log_archive.cfg"):
            os.unlink(os.path.join(nagios_exportdir, f))


def configure_container_creds():
    archive_userinfo = conf.archive_userinfo()
    containerrc = os.path.join(archive_userinfo.pw_dir, "{}.rc".format(service_name()))
    with open(containerrc, "w") as f:
        f.write(conf.container_credentials())
    os.chmod(containerrc, 0o400)
    os.chown(containerrc, archive_userinfo.pw_uid, archive_userinfo.pw_gid)


def configure_sudo():
    logdirs = []
    globs = []
    for d in conf.logdirs():
        isglob = False
        for possible_glob in ["*", "?", "[", "]"]:
            if possible_glob in d:
                isglob = True
                break
        if isglob:
            globs.append(d)
        else:
            logdirs.append(d)

    tmpl_data = {}
    tmpl_data["logdirs"] = logdirs
    tmpl_data["globs"] = globs
    tmpl_data["username"] = conf.archive_username()
    sudo_path = "/etc/sudoers.d/15-{}".format(service_name())
    file_from_template("sudo-log-archive.tmpl", sudo_path, tmpl_data)
    os.chmod(sudo_path, 0o440)
    log("CHARM: Wrote sudoers file {}".format(sudo_path))


def configure_cronjob():
    cronpath = "/etc/cron.d/cla-{}".format(service_name())
    script_dir = os.path.join(charm_dir(), 'files')

    # Get the pre and post expansion cron time strings
    expanded_from_tag = "# Expanded from: "
    expanded_to_tag = "# Expanded to: "
    original_cron_time = ""
    current_cron_time = ""
    # On first run, the log-archive cron file won't exist
    if os.path.isfile(cronpath):
        with open(cronpath, "r") as f:
            for line in f:
                if expanded_from_tag in line:
                    original_cron_time = line.partition(expanded_from_tag)[2].strip()
                elif expanded_to_tag in line:
                    current_cron_time = line.partition(expanded_to_tag)[2].strip()

    logdirs = conf.logdirs()
    if logdirs:
        # Only change the time if the cron_time config value has changed
        # (avoids <<rnd>> expansion changing things every time)
        new_cron_time = conf.cron_time()
        if new_cron_time == original_cron_time:
            cron_time = current_cron_time
        else:
            cron_time = expand_crontimes(new_cron_time)
        delete_after_archive = " -d" if conf.delete_after_archive() else ""
        retention_days = ""
        if conf.retention_days() and int(conf.retention_days()) > 0:
            retention_days = " -p %s" % conf.retention_days()

        cron_cmd = os.path.join(script_dir, "archive-to-{}.py".format(conf.container_type()))
        if conf.container_type() in ['gce', 'azure'] and conf.project() != "":
            cron_cmd += " --project {}".format(conf.project())

        config_file = os.path.join(conf.archive_userinfo().pw_dir, "{}.rc".format(service_name()))
        container_path = conf.container_name() + "/" + conf.container_path()
        preprocessors = ""
        for pp in conf.preprocessors():
            pp = pp.strip()
            # Ensure backwards compatibility with obsolete anonymize_web_log_ips preprocessor.
            if pp == 'anonymize_web_log_ips':
                pp = 'anonymize_web_log'
            if pp == 'anonymize_web_log' and conf.anonymize_web_log_options() != '':
                preprocessors += '--pre "{} {}" '.format(pp, conf.anonymize_web_log_options())
            else:
                preprocessors += '--pre {} '.format(pp)
        cronstr = "%s %s %s -s -f %s -l %s -c %s %s -a %s%s%s\n" % (cron_time, conf.archive_username(), cron_cmd, config_file, " -l ".join(logdirs), container_path, preprocessors, conf.archive_after(), delete_after_archive, retention_days)
        with open(cronpath, "w") as f:
            f.write(juju_header())
            f.write(expanded_from_tag + new_cron_time + "\n")
            f.write(expanded_to_tag + cron_time + "\n")
            f.write(cronstr)
        log("CHARM: Wrote cron file {}".format(cronpath))

    # Ensure we can list log directories
    for logdir in logdirs:
        if not os.path.isdir(logdir):
            continue
        st = os.stat(logdir)
        os.chmod(logdir, st.st_mode | stat.S_IXOTH | stat.S_IROTH)


def configure_nrpe():
    if not conf.nagios_relation_ok() or not conf.log_archive_rel_ok():
        log("CHARM: Nagios or log-archive relation not ready - not configuring nrpe")
        return

    nagios_etcdir = "/etc/nagios/nrpe.d"
    nagios_exportdir = "/var/lib/nagios/export"

    # No nagios dirs probably means no nagios, so don't bother with the rest
    if not (os.path.isdir(nagios_etcdir) and os.path.isdir(nagios_exportdir)):
        log("CHARM: Essential nagios dirs missing - skipping Nagios checks")
        return

    archive_userinfo = conf.archive_userinfo()
    status_file = os.path.join(archive_userinfo.pw_dir, "cla-{}.status".format(service_name()))
    tmpl_data = {}
    tmpl_data["service"] = service_name()
    tmpl_data["status_file"] = status_file
    tmpl_data["script_dir"] = os.path.join(charm_dir(), 'files')

    log("CHARM: Adding check check_log_archive_status")
    checkpath = os.path.join(nagios_etcdir, "check_{}_status.cfg".format(service_name()))
    tmplfile = "nagios-check-log-archive-status.tmpl"
    file_from_template(tmplfile, checkpath, tmpl_data)

    wipe_nrpe_exports(nagios_exportdir)

    configpath = os.path.join(nagios_exportdir,
                              "service__{}_check_{}.cfg".format(conf.nagios_hostname(), service_name()))
    tmpl_data = {}
    tmpl_data["use"] = "active-service"
    tmpl_data["nagios_hostname"] = conf.nagios_hostname()
    tmpl_data["description"] = "State of last {} run on {}".format(service_name(),
                                                                   conf.nagios_hostname())
    tmpl_data["check_command"] = "check_nrpe!check_{}_status".format(service_name())
    tmpl_data["servicegroup"] = conf.nagios_servicegroup()
    tmplfile = "nagios-config.tmpl"
    file_from_template(tmplfile, configpath, tmpl_data)

    # Add a fake log-archive.status file to stop the check failing before the first run
    with open(status_file, "w") as f:
        f.write("DATE: {}\n".format(time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())))
        f.write("Archived: 0 of 0 files\n")

    service_reload("nagios-nrpe-server")


def deconfigure_nrpe():
    log("CHARM: Deconfiguring nrpe")
    nagios_etcdir = "/etc/nagios/nrpe.d"
    nagios_exportdir = "/var/lib/nagios/export"
    checkpath = os.path.join(nagios_etcdir, "check_{}_status.cfg".format(service_name()))
    if os.path.isfile(checkpath):
        os.unlink(checkpath)
    wipe_nrpe_exports(nagios_exportdir)
    service_reload("nagios-nrpe-server")


def remove_legacy_files():
    """ Previous versions of the charm didn't use the service_name in the
        various system files. This removes those older files if they exist
        so that upgrades don't end up with duplicates.
    """
    legacy_paths = [
        os.path.join(conf.archive_userinfo().pw_dir, "containerrc"),
        "/etc/nagios/nrpe.d/check_log_archive_status.cfg",
        "/var/lib/nagios/export/service__{}_check_log_archive.cfg".format(conf.nagios_hostname()),
        "/etc/sudoers.d/10-log-archive",
        "/etc/cron.d/log-archive",
        os.path.join(conf.archive_userinfo().pw_dir, "log-archive.status"),
    ]

    for path in legacy_paths:
        if os.path.exists(path):
            log("CHARM: Removing legacy path {}".format(path))
            os.remove(path)


@hooks.hook("install")
def install():
    log("CHARM: Installing {}".format(service_name()))
    install_packages(conf.container_type())
    archive_userinfo = adduser(conf.archive_username(), system_user=True)
    mkdir(archive_userinfo.pw_dir, owner=conf.archive_username())
    os.chmod(archive_userinfo.pw_dir, 0o755)


@hooks.hook("upgrade-charm")
def upgrade_charm():
    log("CHARM: Upgrading {}".format(service_name()))
    install()


@hooks.hook("config-changed")
def config_changed():
    log("CHARM: Configuring {}".format(service_name()))
    if conf.container_type_changed():
        install_packages(conf.container_type())
    if conf.log_archive_rel_ok():
        configure_container_creds()
        configure_sudo()
        configure_cronjob()
    configure_nrpe()
    remove_legacy_files()


@hooks.hook("juju-info-relation-changed")
@hooks.hook("log-archive-relation-changed")
def log_archive_relation_changed():
    log("CHARM: Changing log-archive relation for {}".format(service_name()))
    config_changed()


@hooks.hook("juju-info-relation-broken")
@hooks.hook("log-archive-relation-broken")
def log_archive_relation_broken():
    log("CHARM: Breaking log-archive relation for {}".format(service_name()))

    # Clear out the saved config to avoid it providing stale data
    if os.path.isfile(conf.config_file()):
        os.unlink(conf.config_file())


@hooks.hook("start")
def start():
    # Nothing to start - just here to keep charm proof happy
    log("CHARM: Starting {}".format(service_name()))


@hooks.hook("stop")
def stop():
    # Nothing to stop - just here to keep charm proof happy
    log("CHARM: Stopping {}".format(service_name()))


@hooks.hook("nrpe-external-master-relation-changed")
def nrpe_external_master_relation_changed():
    log("CHARM: Changing nrpe relation for {}".format(service_name()))
    configure_nrpe()


@hooks.hook("nrpe-external-master-relation-broken")
def nrpe_external_master_relation_broken():
    log("CHARM: Breaking nrpe relation for {}".format(service_name()))
    deconfigure_nrpe()


@hooks.hook("local-monitors-relation-changed")
def local_monitors_relation_changed():
    log("CHARM: Changing monitors relation for {}".format(service_name()))
    configure_nrpe()


@hooks.hook("local-monitors-relation-broken")
def local_monitors_relation_broken():
    log("CHARM: Breaking monitors relation for {}".format(service_name()))
    deconfigure_nrpe()


if __name__ == "__main__":
    hooks.execute(sys.argv)
