#!/usr/bin/python3

import amulet
import base64
import unittest


class TestCharm(unittest.TestCase):
    # config.yaml defaults
    archive_after = '7'
    user = 'log-archive'

    # Test constants
    config_file = '/home/log-archive/log-archive.rc'
    container_name = 'testContainer'
    creds = "user=blah\npassword=foo"
    cron_file = '/etc/cron.d/cla-log-archive'
    log_dir1 = '/var/log/apache2'
    log_dir2 = '/var/log/apt'
    preprocessor = 'gzip'
    sudoers_file = '/etc/sudoers.d/15-log-archive'

    @classmethod
    def configure(cls):
        cls.d = amulet.Deployment(series='xenial')

        # apache2 as the principle charm
        cls.d.add('apache2', 'cs:apache2')
        cls.d.add('log-archive', charm='container-log-archive')
        cls.d.relate('apache2:juju-info', 'log-archive:juju-info')

        # test configuration
        cls.d.configure('log-archive', {
            'container_credentials': base64.b64encode(cls.creds.encode('utf-8')).decode('utf-8'),
            'container_name': cls.container_name,
            'container_type': 'gce',
            'delete_after_archive': 'true',
            'preprocessors': cls.preprocessor,
            'logdirs': '{} {}'.format(cls.log_dir1, cls.log_dir2),
        })

    @classmethod
    def setUpClass(cls):
        cls.configure()

        cls.d.setup(timeout=900)
        cls.d.sentry.wait()

        cls.cla = cls.d.sentry['log-archive'][0]

    def test_cronjob(self):
        """ Verify the logdirs, cronjob and archive_after_default are setup.
        """
        expected_config = [
            'archive-to-',
            '-l {}'.format(self.log_dir1),
            '-l {}'.format(self.log_dir2),
            '-c {}'.format(self.container_name),
            self.preprocessor,
            '-a {}'.format(self.archive_after),
            '-d',  # delete_after setting
        ]

        cronjob = self.cla.file_contents(self.cron_file)
        for conf in expected_config:
            self.assertTrue(cronjob.find(conf) != -1)

    def test_credentials(self):
        """ Verify the credentials and container are correctly set.
        """
        stat = self.cla.file_stat(self.config_file)
        self.assertEqual(stat['mode'], '0o100400')

        self.assertEqual(self.creds, self.cla.file_contents(self.config_file))

    def test_sudoers(self):
        """ Verify the sudoers file is correctly setup.
        """
        stat = self.cla.file_stat(self.sudoers_file)
        self.assertEqual(stat['mode'], '0o100440')

        _, exit_code = self.cla.run('visudo -c -f {}'.format(self.sudoers_file))
        self.assertEqual(exit_code, 0)

    def test_user(self):
        """ Verify the log archive user was created.
        """
        _, exit_code = self.cla.run('getent passwd {}'.format(self.user))
        self.assertEqual(exit_code, 0)


if __name__ == "__main__":
    unittest.main()
